/* Macro description */
  /*------------------------------------------------------------------*
   | The documentation and code below is supplied by the Methodology Center, PSU.             
   |              
   *------------------------------------------------------------------*/
                                                                                      
  /*------------------------------------------------------------------*
   | MACRO NAME  : %TVEM_zip
   | SHORT DESC  : Fit time-varying Coefficient Model to intensive longitudinal data 
   | for count responnse using zero-inflated Poisson (ZIP) model
   *------------------------------------------------------------------*
   | CREATED BY  : Runze LI, Xianming TAN, Jingyun YANG and Liying Huang (04/02/2014)
   *------------------------------------------------------------------*
   |
   |
   |  Example: Suppose the dataset simulated_poisson_data.sas7bdat has the following variables: 
   |		subj: subjectsEidentification variable; 
   |		t:    measurement time for each observation, different subjects may have different measurement times;
   |		y:    the dependent variable;
   |		x0:   the intercept variable which equals to 1 for all observations;  
   |		x1:   a binary covariate;
   |		x2:   a continuous covariate.   
   |	
   |	Some observations in this data set are as follows:
   |	
   |
   |				subj   y      t      X0     X1         X2                  
   |				-------------------------------------------------
   |				1      4    0.00591     1     0    -1.71718
   |				1      8    0.01378     1     0    -1.99814
   |				1      0    0.01772     1     1     0.59689
   |				1      3    0.02559     1     1    -0.21331
   |				1      3    0.02953     1     1     1.10735
   |				2      1    0.00197     1     1     1.28186
   |				2      0    0.00591     1     1     1.55535
   |				2      5    0.00984     1     0    -0.84808
   |				2      2    0.01378     1     1     0.80863
   |				2      2    0.02165     1     0     0.60380
   |				.....................................
   |
   |		
   |
   |   We want to fit a time-varying coefficient model in which y is the dependent variable (count response but zero inflated), 
   |   x0, x1, x2 are the covariates. The coefficient of x0, x1 and x2 may vary over time.
   |   When there are excessive zeros in the response, we could fit a zero-inflated Poisson model.
   |   In this example, assume that x1 is the zero-inflated covarite.
   |   We could call the Macro as follows:
   |
   |	%TVEM_ZIP(  	
   |		mydata = simulated_data,  
   |		id = subj,  		
   |		time = t, 		
   |		dep = y,    		
   |		tcov = x0 x1 x2,  	  
   |		cov_knots = 6 6 6,
   |        cov_zip=x1
   |		);
   |
   |	where cov_knots determines the number of B-spline basis functions used to approximate the corresponding 
   |	coefficeint function, see User's guide for detailed description. 
   |
   |  	References:  User's guide
   |
   *------------------------------------------------------------------*
   | OPERATING SYSTEM COMPATIBILITY
   |
   | UNIX SAS v8   :   
   | UNIX SAS v9   :
   | MVS SAS v8    :
   | MVS SAS v9    :
   | PC SAS v8     :	
   | PC SAS v9     :	Yes
   *------------------------------------------------------------------*
   | Copyright 2011 The Methodology Center, Penn State University.
   |
   | This program is free software; you can redistribute it and/or
   | modify it under the terms of the GNU General Public License as
   | published by the Free Software Foundation; either version 2 of
   | the License, or (at your option) any later version.
   |
   | This program is distributed in the hope that it will be useful,
   | but WITHOUT ANY WARRANTY; without even the implied warranty of
   | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   | General Public License for more details.
   *------------------------------------------------------------------*/

%MACRO TVEM_ZIP(mydata = ,  
		id = ,  			/* the varaible indicating subjects' identification */
		time = , 			/* the varaible contains the measurement times */
		dep = ,    	/*  Y (outcome) */
		cov = ,		/* covariates with constant coefficient*/
		class_var = ,     /* classifcation varaibles */
		tcov = ,  	/* Intercept term should be specified, I may also  add dwkave */  
		cov_knots = ,      /* match the number of covaraites in TCOVs*/
		deg = 3,  		/* degree of B-spline, default is cubic spline, deg=3 */
		evenly = 0,   	/* if 1, knots are evenly in space, otherwise(e.g. 0 ), knots are evenly on quantiles */
		scale = 100,				 /* determine the number of fineness of plots valid only when Plot=yes */
		outfilename = , 				/* file name for plot_data.csv */
		cov_zip= 
		);

%DO;
%macro grabpath;
%qsubstr(%sysget(SAS_EXECFILEPATH),1,%length(%sysget(SAS_EXECFILEPATH))-%length(%sysget(SAS_EXECFILEname)))
%mend grabpath;

%PoissonMOVEBspline(  
		&mydata ,  
		&id ,  			
		&time , 		
		&dep  ,    		
		&cov  ,			
		&class_var  ,   
		&tcov  ,  		 
		&cov_knots  ,
		&deg ,  			
		&evenly ,   	
		&scale ,	
		&outfilename ,
		&cov_zip	
	);
%END;

%MEND TVEM_ZIP;

%MACRO PoissonMOVEBspline(  		/* implement TVEM for ZIP using B-spline*/
		mydata  ,  
		id  ,  			/* the varaible indicating subjects' identification */
		time  , 			/* the varaible contains the measurement times */
		dep  ,    	/*  Y (outcome) */
		cov  ,		/* covariates with constant coefficient*/
		class_var  ,     /* classifcation varaibles */
		tcov  ,  	/* Intercept term should be specified, I may also  add dwkave */  
		cov_knots  ,      /* match the number of covaraites in TCOVs*/
		deg ,  		/* degree of B-spline, default is cubic spline, deg=3 */
		evenly ,   	/* if 1, knots are evenly in space, otherwise(e.g. 0 ), knots are evenly on quantiles */
		scale ,				 /* determine the number of fineness of plots valid only when Plot=yes */
		outfilename ,  				/* file name for plot_data.csv */
		cov_zip
	);

/*step 1: data separation */
	DATA _t_tcov;   /* covariates with time-varying coefficients */
		SET &mydata;
		WHERE &time >. ;
		KEEP &tcov; 
	RUN;

	DATA _id_dep_cov; /*id, dependent varaible, covariates with constant coefficients, &time may be needed for R specification */
		SET &mydata;
		WHERE &time >. ;
		KEEP &id &dep &cov &time &cov_zip;   /* A new line with zip variable included*/ 
	RUN;

	DATA _t_orig;   /* covariates with time-varying coefficients */
		SET &mydata;
		WHERE &time >. ;
		KEEP &time; 
	RUN;

	PROC SORT DATA=_t_orig OUT=_t_sorted;  /*sorted time variable, needed for inner knots calculation */ 
		BY &time;
	RUN;

	PROC SORT DATA=_t_orig OUT=_t_distinct NODUPKEY;  /*sorted time variable, needed for inner knots calculation */ 
		BY &time;
	RUN;

	DATA _t_orig;   /* Keep _idx to order the observations, be consistent with _id_dep_cov  */
		RETAIN _idx 0;
		SET _t_orig;
		_idx = _idx + 1;
	RUN;

/*step 1: Generate B-spline transformation for &TCOV */

PROC IML;
	/* Part 1: three functions, my_Bspline(...), my_vec_Bspline(...), and my_knots(...) */
	/* Please simply skip this part if you'are not interested in technique details */

	START my_Bspline(x, knots, d);
	/*	evaluate the values of (num of knots) + d +1 B-spline basis functions at x, x is a real number*/
	/*	output a vector of (num of knots) + 1 - d real numbers */
	/*    knots include both exterior and inner knots */
		n_ext = ncol(knots);
		all_knots = (1:(n_ext+2));
		all_knots[1] = knots[1] - (1e-12);
		all_knots[2:(n_ext+1)]= knots;
		all_knots[n_ext+2] = knots[n_ext] + (1e-12); 

		n_ext =n_ext+2;

		tmp=0.0*(1:(n_ext-1));

		DO i = 1 TO (n_ext-1);
			IF (x>=all_knots[i] & x<all_knots[i+1]) THEN tmp[i]=1.0; 
		END; 

		j=1;
		DO WHILE (j<=d); /* the De Boor formula, refer to Eilers & Marx (1996) */
			DO i = 1 TO (n_ext-j-1); 
				w1 = (x-all_knots[i])/(all_knots[i+j] - all_knots[i]); 
				w2 = (all_knots[i+j+1]-x)/(all_knots[i+j+1] - all_knots[i+1]); 
				tmp[i] = w1*tmp[i] +w2*tmp[i+1];
			END;
			j=j+1;
		END;	
		RETURN(tmp[1:(n_ext-d-1)]);
	FINISH;

	START my_vec_Bspline(xx, knots, d);
	/*	evaluate the values of the (num of knots) + d +1 B-spline basis functions at xx */ 
	/*	xx is a vector of real numbers*/
	/*	output a real matrix with dim(xx) rows,  and (num of knots) + d +1 columns */
		n_row = ncol(xx);
		n_col = ncol(knots)+2-d-1;		
		out = J(n_row, n_col, .);  
		DO i=1 TO n_row; 
			out[i, ] = t(my_Bspline(xx[i], knots, d));
		END;			
		RETURN(out);
	FINISH;

	START my_knots(samp, nknots, evenly, d);  
	/* Calculate nknots inner knots and add 2*d exterior knots */
	/* the inner knots are uniform in space (evenly=1) or uniformly on quantiles (evenly != 1) */
	/* samp should be sorted ascendingly, i.e., from small to large */

		len = ncol(samp);
		*len = max(nrow(samp), );
		all_knots = 0*(1:(nknots+2*d));
		all_knots[1:d] = samp[1]-(1e-12)*(d:1);
		all_knots[(nknots+d+1):(nknots+2*d)] = samp[len]+(1e-12)*(1:d);

		IF (evenly=1) THEN 
				DO;
					dist  = (samp[len]-samp[1])/(nknots+1); 		
					all_knots[(d+1):(d+nknots)] = samp[1]+dist*(1:nknots);
				END; 
			ELSE DO; 
				*inn_knots = quantile(samp, p=(1:nknots)/(nknots+1)); 
				* According to Bspline (PROC TRANSREG), quantiles should be on data points, no interpolation;
 					j=1; 
					DO i = 1 TO nknots;
						DO WHILE ( (j-1)/(len-1) < i/(nknots+1) );
							j=j+1;
						END;
						all_knots[d+i] = samp[j];
					END;
				END;
		RETURN(all_knots);
	FINISH;

	/* Part 2: generate Bspline matrix, given t_cov and options for knots */
	/* 		and read data into matrix */

	USE _t_sorted;
	READ ALL INTO mat_t_sorted; /* &time only, sorted */
	USE _t_distinct;
	READ ALL INTO mat_t_distinct; /* &time only, distinct time points only */

	n_pt = &scale -1;
	mat_t_dense = t( mat_t_sorted[1]+ ((0:n_pt)/n_pt)*( mat_t_sorted[ nrow(mat_t_sorted) ] - mat_t_sorted[1] ) );

	k_s = {&cov_knots}; /* a vector of number of inner knots, one for each covariate in &TCOV */
	
	DO i = 1 TO ncol(k_s);
		all_knots = my_knots(t(mat_t_sorted), k_s[i], &evenly, &deg);
		mat_t_splines = mat_t_splines || ( my_vec_Bspline( t(mat_t_distinct), all_knots, &deg) ); /* distinct time points */
		mat_dt_splines = mat_dt_splines || ( my_vec_Bspline( t(mat_t_dense), all_knots, &deg) ); /* uniform distributed time points */
	END;

	mat_t_splines = (mat_t_distinct || mat_t_splines);

	CREATE _t_spline FROM mat_t_splines[colname= "_time"]; /* _time col2 - col( 1+ sum( k_s[i] + &deg + 1 )  )*/
	APPEND FROM mat_t_splines; 

	/* output mat_dt_splines to dataset dt_spline
		dataset t_dense is useful for plotting  */


	CREATE _dt_spline FROM mat_dt_splines; /* col1 - col( sum( k_s[i] + &deg + 1 )  )*/
	APPEND FROM mat_dt_splines;
	CREATE _t_dense FROM mat_t_dense [colname= "&time"]; /* denser time points */
	APPEND FROM mat_t_dense;
 QUIT;

PROC SQL;
	CREATE TABLE _t_orig_spline AS
	SELECT A.*, B.*  
	FROM _t_orig A, _t_spline B 
	WHERE A.&time = B._time
	ORDER BY _idx;

PROC IML;
	k_s = {&cov_knots}; /* a vector of number of inner knots, one for each covariate in &TCOV */
	j = k_s[+] + (&deg+1)*ncol(k_s);
	c_name = ( compress(  CATT( "col", char(j+1, 8, 0)) ) ); 

	USE _t_orig_spline;
	READ ALL VAR("col2":c_name) INTO mat_tspline;  /* &time &tcov, unsorted,*/ 

	USE _t_tcov;
	READ ALL VAR{&tcov} INTO mat_tcov;  /* &time &tcov, unsorted,*/ 
							/*VAR{&tcov} makes sure the order of columns in mat_tcov is consistent with &tcov(and hence cov_knots)*/

	id_u = 0; 
	DO i = 1 TO ncol(k_s);
		id_l = id_u + 1; id_u = id_l + (k_s[i]+&deg+1)-1;
		mat_t_splines = mat_t_splines || ( (mat_tspline[, id_l: id_u] ) # mat_tcov[,i]);
	END;

	/* output mat_Tcov_splines to dataset tcov_spline 
				output mat_dt_splines to dataset dt_spline */
	CREATE _tcov_spline FROM mat_t_splines; /* col1 - col( sum( k_s[i] + &deg + 1 )  )*/
	APPEND FROM mat_t_splines; 
QUIT;

/* Step 3: Proc Mixed */
/* results of this step are: my_Fix_est my_CovB */

	DATA model_data;  /* merge  */
		MERGE _id_dep_cov _tcov_spline;
	RUN;

	PROC CONTENTS DATA=_tcov_spline OUT=_tmp NOPRINT;	RUN;
	PROC SQL; 		/* find out the number of varaibles reduced from B-splines */
		RESET NOPRINT;
		SELECT (MAX(VARNUM)) INTO :aa FROM _tmp;  
	QUIT;
	%LET size = %TRIM(&aa);


	*ODS SELECT  ClassLevels Nobs OverallANOVA  FitStatistics  ParameterEstimates; /* selected outputs*/
	*ODS OUTPUT FitStatistics=_my_fit ParameterEstimates=_my_Fix_est InvXPX=_my_CovB NOBs=_my_Nobs;
	*ODS SELECT  ClassLevels Nobs  ModelFit  ParameterEstimates; /* selected outputs*/
/* 
	For zero-inflated Poissson Regression
*/

/* 
	For zero-inflated Poissson Regression
*/
ODS SELECT ALL;
ODS OUTPUT ModelFit=_my_fit ParameterEstimates=_my_Fix_est NOBs=_my_Nobs CovB=_my_CovB;
	PROC GENMOD DATA=model_data; 
		CLASS &id &class_var;
		MODEL &dep = col1-col&size &cov /covb dist=zip noint type3;
        zeromodel &cov_zip/link=logit;
	RUN;

/*
	PROC CONTENTS DATA=_my_covB OUT=_tmp NOPRINT;	RUN;
	PROC SQL; 		/* find out the number of varaibles reduced from B-splines */
*		RESET NOPRINT;
*		SELECT (MAX(VARNUM)) INTO :aa FROM _tmp;  
*	QUIT;
*	%LET p = %TRIM(&aa); * number of coeffieicnt parameters;
*/;

DATA _my_Nobs;
	SET _my_Nobs;
	WHERE Label='Number of Observations Read'; 
	KEEP NObsUsed;
RUN;

/*
	DATA _est;
		MERGE _my_fit _my_Nobs;
	RUN;
	DATA _est;
		SET _est;
		_P_ = &p - 2;
		 q  = NObsUsed/(NObsUsed- _p_);
		_2LLK_= NObsUsed*log(RootMSE*RootMSE/q);
		_AIC_ = _2LLK_ + 2*_p_ ;
		_BIC_ = _2LLK_ + 2*(_p_ + 2) - 2*q*q; 
		label _P_ = 'Number of parameters';
		label _2LLK_='-2 Res Log Likelihood';
		label _AIC_ = ' Akaike information criterion ';
		label _BIC_ = ' Sawa Bayesian information criterion';
		KEEP _P_ NObsUsed _2LLK_ _AIC_ _BIC_  ;
	RUN;

PROC TRANSPOSE DATA=_est OUT=_T_est;
RUN; 

DATA _T_est;
	SET _T_est;
	DROP _name_;
RUN;
*/
PROC PRINT DATA=_my_fit noobs label; 
RUN;


/*step 4: calculate curves and CI's */
  PROC IML;
	USE _dt_spline;
	READ ALL INTO mat_dt_spline;  /* Spline basis functions at denser time points */

 	USE _My_fix_est;
	read ALL VAR {estimate} INTO mat_fix_est where(parameter^="Intercept");

	k_s = {&cov_knots}; /* a vector of number of inner knots, one for each covariate in &TCOV */
	j =1+ k_s[+] + (&deg+1)*ncol(k_s);  /*need to double checkw with Xianming*/
	c_name = ( compress(  CATT( "prm", char(j, 8, 0)) ) ); 

	USE _My_fit;
	READ ALL VAR("ValueDF") INTO MSE where(criterion=:"Deviance");

	USE _My_CovB;
	READ ALL VAR("prm2":c_name) INTO mat_covB;  /* only read volumns relevant to B-spline basis */
    
	*mat_covB = mat_covB*(sqrt(MSE));  /* \sigma^2 * Inv(X'X) */

	*k_s = {&cov_knots}; /* vector of B-spline orders */ 
	/* ncol(k_s) = number of covariates with time-varying coefficients */
	/* T = mat_t_dense */

	idx_u=0;  /* index sets for picking up suitable columns */
	DO i = 1 TO ncol(k_s);
		idx_l=idx_u+1;
		idx_u = idx_l + (k_s[i]+&deg+1) -1 ;  /* for P-spline, 3 will be replaced by a parameter */

		/* values at denser points */
		Coef = mat_dt_spline[, idx_l:idx_u]*mat_fix_est[idx_l:idx_u];
		/* pointwise variances */
		cov_Coef = vecdiag( ( mat_dt_spline[, idx_l:idx_u]*mat_covB[idx_l:idx_u, (idx_l):(idx_u)])*t(mat_dt_spline[, idx_l:idx_u]) );
		IF ( MIN(cov_Coef)<0 ) THEN cov_Coef[loc(cov_Coef<0)] = .;  /* if <0, then missing*/

		/* calculate CIs */ 
		Coef_L = Coef - 1.96*sqrt(cov_Coef);
		Coef_U = Coef + 1.96*sqrt(cov_Coef);
        /*
		Coef_OR=exp(Coef);
		Coef_L = exp(Coef - 1.96*sqrt(cov_Coef));
		Coef_U = exp(Coef + 1.96*sqrt(cov_Coef));
		*/
		mat_coef = mat_coef || ( (coef_L||coef) || coef_U);

		var_name = SCAN("&tcov", i);
		var_names = ( catt(var_name, "_L") || var_name || catt(var_name,"_U") );
		cname = cname || var_names;
	END;
	
	CREATE _coef FROM mat_coef[colname=cname]; /* denser time points */
	APPEND FROM mat_coef;
	CREATE _covb FROM mat_covb; /* denser time points */
	APPEND FROM mat_covb;
  QUIT;

/*step 5: Plot Curves and Confidence Bands */

	PROC CONTENTS DATA=_coef OUT=_tmp NOPRINT;	RUN;
	PROC SQL;
	/* find out the number of varaibles, reduced from B-splines */
		RESET NOPRINT;
		SELECT (MAX(VARNUM))/3 INTO :aa FROM _tmp;  
	QUIT;
	%LET size = %TRIM(&aa);

	DATA plot_data;
		MERGE _t_dense _coef;
	RUN;

	%DO i=1 %TO &size;
		%LET var_name = %SCAN(&tcov, &i);
		/*
		TITLE1 "Coefficient and 95% Confidence Band for Covariate: &var_name";
		AXIS  
			LABEL=('Coefficient' justify=right "of &var_name")
		      WIDTH=3;
		LEGEND LABEL =none value=('Lower Bound' 'curve' 'Upper Bound') POSITION=(bottom left inside) MODE=share DOWN = 2; 
		SYMBOL1 
			COLOR=green
			VALUE=point
			INTERPOL=join
        		LINE=2
        		WIDTH=3;
		SYMBOL2 
      		COLOR=red
			VALUE=point
			INTERPOL=join
			LINE=1
	        	WIDTH=3;
		SYMBOL3 
			COLOR=green
			VALUE=point
			INTERPOL=join
			LINE=2
        		WIDTH=3;

		PROC GPLOT DATA=plot_data;
			PLOT (&var_name._L &var_name &var_name._U )* &time / overlay legend = legend vaxis=axis;
		RUN;
	%END;

	TITLE1;
	SYMBOL1; 
	SYMBOL2; 
	SYMBOL3; 
	AXIS;
	LEGEND;
*/

proc template;                                                                                                                          
  define statgraph mygraph_tvem;                                                                                                             
  begingraph;                                                                                                                           
  entrytitle "Coefficient Curve and 95% Confidence Band for Covariate: &var_name";                                                                                                             
    layout lattice / columndatarange=union rowweights=(0.9 0.1);                                                                        
                                                                                                                                        
/* Overlay a Bandplot and a Seriesplot */                                                                                               
      layout overlay /yaxisopts=(label="Coefficient of &var_name");                                                                               
        bandplot x=&time limitupper=&var_name._U limitlower=&var_name._L / datatransparency=.3 modelname="fit" 
           legendlabel="95% Confidence Limits";                                                         
        seriesplot x=&time y=&var_name;  
		referenceline y=0/LINEATTRS=(pattern=4); 
      endlayout;                                                                                                                        
                                                                                                                                        
/* Draw the Month values beneath the graph */ 
/* 
      blockplot x=day block=month / display=(values label);   
*/ 
    endlayout;                                                                                                                         
  endgraph;                                                                                                                             
  end;                                                                                                                                  
run;                                                                                                                                    
                                                                                                                                        
proc sgrender data=plot_data template=mygraph_tvem;                                                                                               
run;

%END;


/*step 6: delete temporary data sets  */

PROC DATASETS;
	DELETE _: ;
RUN; 
QUIT;

%LET t_dir = %grabpath; *%sysfunc(getoption(work));

%IF %nrbquote(&outfilename) =  %THEN %DO; 
	%LET outfilename = &t_dir.\plot_data.csv;
%END;

PROC EXPORT DATA= WORK.Plot_data 
            OUTFILE= "&outfilename" 
            DBMS=CSV REPLACE;
RUN;


%MEND PoissonMOVEBspline;


